"Prediction Module"
import logging
from multiprocessing import process
import os

import numpy as np
from numpy import array
import rasterio
from rasterio.windows import Window

from classifier.utils.raster import iterate_rasters_for_data
from classifier.utils.general import write_tifs
from classifier.settings import US_ALGORITHMS
from classifier.timeseries import get_timeseries_samples

# need to set the temp dir for multiprocessing in a container
process.current_process()._config['tempdir'] = '/tmp/'

# Load all rasters in a windowed manner (in a vrt) and do the prediction
PREDICT_LOGGER = logging.getLogger(__name__)


def prediction_class_probabilities(predicted, valid, meta, config_dict, window):
    """Gets probability for the classified class and writes the tif for a
    windows

    Args:
        predicted (np.ndarray): The probabilities for all classes
        valid (np.ndarray): Array with valid values for prediction
        meta (dict): rasterio meta for writing tif
        window: rasterio.Window to predict and write

    """
    labels_internal_proba = (np.nanmax(predicted, axis=1)
                             .astype(np.float32))
    result_proba = np.full(valid.shape, -9999, np.float32)
    result_proba[valid] = labels_internal_proba
    write_tifs(os.path.join(config_dict['tmp_dir'], 'probability'),
               window,
               meta['dst_proba_meta'],
               result_proba)

def prediction_all_probabilities(
        predicted: array,
        valid: array,
        labels: dict,
        meta: dict,
        config_dict: dict,
        window: Window):
    """Gets probability for the classified class and writes the tif for a
    windows

    Args:
        predicted (np.ndarray): The probabilities for all classes
        valid (np.ndarray): Array with valid values for prediction
        labels (dict): Conversion dictionary for labels
        meta (dict): rasterio meta for writing tif
        window (rasterio.Window): processing window

    """
    # loop over the classes and save them all separately
    for internal_label in range(predicted.shape[1]):
        probabilities = predicted[:, internal_label]
        result_proba = np.full(valid.shape, -9999, np.float32)
        result_proba[valid] = probabilities
        write_tifs(
            os.path.join(config_dict['tmp_dir'], str(labels[internal_label])),
            window,
            meta['dst_proba_meta'],
            result_proba)

def do_prediction_proba_su(data, valid, model_dict, config_dict, meta, window):
    """Does the actual prediction and probability description for supervised
    methods

    Args:
        data: Numpy array with features
        valid: Which parts of the numpy array have valid data
            model_dict:  The model dictionary containting algorithm info
            config_dict: The cli configuration
            meta: The rasterio meta dictionary for the outputs
            window: The rasterio Window for which to do the prediction

    Returns:
        The predictions array

    """
    predicted = model_dict['model'].predict_proba(data)
    # Get the final indices of the highest probabilities
    labels_external = np.nanargmax(predicted, axis=1)

    # Label keys to int
    model_dict['labels'] = {
        int(k): int(v) for k, v in
        model_dict['labels'].items()
    }
    # Indices need to be transformed to the proper labels
    labels_transformed = np.vectorize(
        model_dict['labels'].get)(labels_external)
    result = np.full(valid.shape, -9999, np.int32)
    result[valid] = labels_transformed

    # See if we need to output some of the modelled probabilities
    if config_dict['su_probability']:
        prediction_class_probabilities(
            predicted,
            valid,
            meta,
            config_dict,
            window
        )
    if config_dict['su_all_probabilities']:
        prediction_all_probabilities(
            predicted,
            valid,
            model_dict['labels'],
            meta,
            config_dict,
            window
        )

    return result

def do_prediction_proba(data, valid, model_dict, config_dict, meta, window):
    """Calls the actual probability functions for (un)supervised
    methods

    Args:
        data: Numpy array with features
        valid: Which parts of the numpy array have valid data
            model_dict:  The model dictionary containting algorithm info
            config_dict: The cli configuration
            meta: The rasterio meta dictionary for the outputs
            window: The rasterio Window for which to do the prediction

    Returns:
        The probability array

    """
    if config_dict['app_algorithm'] == 'singleclass':
        result = predict_single_class(
            data,
            valid,
            model_dict,
            config_dict,
            meta,
            window
        )
        return result
    return do_prediction_proba_su(data,
                                  valid,
                                  model_dict,
                                  config_dict,
                                  meta,
                                  window)


def gather_data_for_prediction(rasters, meta, config_dict, window):
    """ Gather all the data from the input rasters and return validity raster

    Args:
        prediction_dict (dict): parameters for the prediction
        window (rasterio Window): window of the chunk to predict

    Returns:
        data (np.ndarray): Array with data for prediction
        valid (np.ndarray): Validity array
    """
    with rasterio.open(rasters[0], 'r') as first:
        f_data = first.read(window=window)
        # Set the correct sizes for output blocks and set the transformation
        for meta_type in ['dst_meta', 'dst_proba_meta']:
            meta[meta_type].update(
                width=f_data.shape[2],
                height=f_data.shape[1],
                transform=first.window_transform(window))

        # Create the data array
        data_array = np.zeros((meta['bandcount'],
                               f_data.shape[1],
                               f_data.shape[2]))

        data, valid = iterate_rasters_for_data(
            window,
            first,
            f_data,
            data_array,
            rasters,
            config_dict
        )
        return data, valid

def gather_data_for_prediction_ts(
        rasters: list,
        meta: dict,
        config_dict:dict,
        window: Window):
    """Gathers all the data neccessary for a prediction using a raster TS

        Args:
            prediction_dict(dict): raster dates and paths
            window(rasterio.Window): window to get data for
        Returns:
            data (np.array): gathered and imputed data
            valid (np.array): Array with indices of locations of valid data
     """
    # Get all data in a df with pixels as columns and dates and bands as the
    # indices
    with rasterio.open(rasters[0], 'r') as src:
        shape = src.read(window=window).shape
        x_min, y_min, x_max, y_max = src.window_bounds(window)
        for meta_type in ['dst_meta', 'dst_proba_meta']:
            meta[meta_type].update(
                width=shape[2],
                height=shape[1],
                transform=src.window_transform(window))

    window_polygon = [[(x_min, y_min),
                       (x_min, y_max),
                       (x_max, y_max),
                       (x_max, y_min)]]

    rois = {'properties': {'id': 1},
            'geometry': {'coordinates': window_polygon,
                         'type':'Polygon'}}

    pixel_data = get_timeseries_samples(rasters,
                                        rois,
                                        out_dir=None,
                                        config_dict=config_dict)
    data = pixel_data.transpose().values
    # All TS is imputed, so we assume that all data is valid.
    valid = np.ones(shape[1:], dtype=np.bool)
    return data, valid


def prediction(window, rasters, model_dict, meta, config_dict):
    """ Prediction for a window and save to a tmp file

    Args:
        window A rasterio window.Window
        rasters  The input raster list
        model_dict Dictionary specifying the algorithm, model labels
        meta The meta files from source, destination and dest_proba
        config_dict Command line arguments
    """
    if config_dict['app_rasters_are_timeseries']:
        data, valid = \
            gather_data_for_prediction_ts(rasters, meta, config_dict,  window)
    else:
        data, valid = \
            gather_data_for_prediction(rasters, meta, config_dict, window)
    if not valid.any():
        return

    if not config_dict['app_algorithm'] in \
           US_ALGORITHMS:
        result = do_prediction_proba(
            data,
            valid,
            model_dict,
            config_dict,
            meta,
            window
        )
    else:
        labels_external = model_dict['model'].predict(data)
        result = np.full(valid.shape, -9999, np.int32)
        result[valid] = labels_external

    # Write the classification results to tifs
    write_tifs(os.path.join(config_dict['tmp_dir'], 'classification'),
               window,
               meta['dst_meta'],
               result)


def predict_proba_single_class(data,
                               valid,
                               model_dict,
                               config_dict,
                               meta,
                               window):
    """
    Predict the probability for an array
    Args:
        data: input data (array)
        valid: where input data is valid (array)
        model_dict: Model dictionary containing algorithm info
        meta: rasterio output meta dictionary
        window: rasterio window

    Returns:
        predicted probability (array
        labels (list)
        meta: rasterio meta for output
    """
    meta.update(dtype=rasterio.dtypes.float32, nodata=-9999)
    result = np.full(valid.shape, -9999, np.float32)
    labels_internal_highest_p = model_dict['model'].decision_function(data)
    result_proba = result.copy().astype(np.float32)
    result_proba[valid] = labels_internal_highest_p

    write_tifs(os.path.join(config_dict['tmp_dir'], 'probability'),
               window,
               meta,
               result_proba)


def predict_single_class(data,
                         valid,
                         model_dict,
                         config_dict,
                         meta,
                         window):
    """
    Predict the probability for an array
    Args:
        data: input data (array)
        valid: where input data is valid (array)
        model_dict: Model dictionary containing algorithm info
        config_dict: Configuration dictionary
        meta: rasterio output meta dictionary
        window: rasterio window

    Returns:
        predicted array

    """
    # Get the final classification of the highest probabilities
    predicted = model_dict['model'].predict(data)
    # print(np.unique(predicted))
    result = np.full(valid.shape, -9999, np.int32)

    # See if we need to output some of the modelled probabilities
    if config_dict['su_probability']:
        predict_proba_single_class(data,
                                   valid,
                                   model_dict,
                                   config_dict,
                                   meta,
                                   window)
    result[valid] = predicted
    return result



"""Raster Utilities"""
import itertools
import logging
import os
import re
import subprocess
import sys
import numpy as np
import rasterio
from rasterio.enums import Resampling
from rasterio.vrt import WarpedVRT

from classifier.utils.general import impute


UTILS_RASTER_LOGGER = logging.getLogger(__name__)


def stitch_(out_ds, temp_dir, meta, dtype='Int32'):
    """
    Stitch all tifs together to a single tif
    Args:
        out_ds: output file
        temp_dir: The temp directory containing the tifs
        meta: rasterio meta dictionary for output
        dtype: the raster's dtype

    """
    filelist = [os.path.join(temp_dir, x) for x in os.listdir(temp_dir) if
                x.endswith('tif')]
    input_list = temp_dir+'input_list.txt'
    with open(input_list, 'w', encoding='UTF-8') as tif_list:
        for files in filelist:
            tif_list.write(f'{files}\n')
    cmd_pred = (
        f"""gdalbuildvrt -srcnodata {meta['dst_meta']['nodata']} \
        {out_ds}.vrt \
        -q \
        -input_file_list {input_list}
        """
    )
    subprocess.call(cmd_pred, shell=True)
    cmd_pred = (
        f"""
        gdalwarp -ot {dtype} -q -multi -wo "NUM_THREADS=ALL_CPUS" -co \
        "TILED=YES"  -co "COMPRESS=DEFLATE" -co "BIGTIFF=YES" -overwrite \
        -srcnodata {meta['dst_meta']['nodata']} {out_ds}.vrt {out_ds}.tif
        """
    )
    subprocess.call(cmd_pred, shell=True)
    os.remove(out_ds+'.vrt')


def stitch(out_prefix, meta, config_dict):
    """
     Stitch the temporary files together
    Args:
        out_prefix (str): The output directory
        meta (dict): The rasterio meta dicts from source, destination and
                     dest_proba


    """
    # Find all directories that have chunks.
    dirs = os.listdir(config_dict['tmp_dir'])
    for directories in dirs:
        if directories.startswith('class'):
            out_file = os.path.join(out_prefix, 'classification')
            dtype = 'Int32'
        elif directories.startswith('probab'):
            out_file = os.path.join(out_prefix, 'probability')
            dtype = 'Float32'
        else:
            out_file = os.path.join(out_prefix, directories)
            dtype = 'Float32'
        UTILS_RASTER_LOGGER.info("Now stitching %s", out_file)
        stitch_(out_file,
                os.path.join(config_dict['tmp_dir'], directories),
                meta,
                dtype)



def iterate_rasters_for_data(
        window, first, f_data,
        data_array, rasters, config_dict):
    """
    Iterate over all rasters to add data into a single array
    Args:
        window: rasterio window
        first: first raster in the list
        f_data: the data from the first raster
        data_array: the array where all raster data is aggregated
        rasters: list of rasters


    Returns:
        The gathered data (array) and array where this data array is valid

    """
    bcounter = first.count
    data_array[:bcounter] = f_data
    # If there is more than 1 raster
    if len(rasters) > 1:

        # Create a vrt with the window and add the data to the array
        for raster in rasters[1:]:
            with rasterio.open(raster, 'r') as src:
                with WarpedVRT(src, resampling=Resampling.bilinear,
                               width=first.width,
                               height=first.height,
                               transform=first.transform,
                               crs=first.crs) as vrt:
                    raster_data = vrt.read(window=window)
                    # Fill the array with this data
                    data_array[
                        bcounter:bcounter + vrt.count,
                        :raster_data.shape[1],
                        :raster_data.shape[2]] = \
                        raster_data[:, :data_array.shape[1],
                                    :data_array.shape[2]]
                    bcounter += vrt.count
    data_array[data_array == first.nodatavals[0]] = np.nan
    if isinstance(data_array, list):
        data_array = np.concatenate(data_array, axis=0)
    if config_dict['app_imputation'] and data_array.size > 0:
        data_array = impute(data_array, config_dict)
    valid = np.isfinite(data_array).all(axis=0)
    data = data_array[:, valid].T
    return data, valid


def get_meta(rasters, window_size=1024):
    """ Get the metadata from the input file and create the windows list to
    iterate through

    Args:
        raster  input raster
        window_size The size of the windows that will be processed
            """

    first_raster = rasters[0]
    # Read the metadata of the first and return metadata for dst and dst_proba
    with rasterio.open(first_raster, 'r') as first:
        f_meta = first.meta
        meta = f_meta.copy()
        meta.update(dtype=rasterio.dtypes.int32, nodata=-9999, count=1,
                    driver='GTiff')
        # Create the windows
        col_offs = np.arange(0, first.width, window_size)
        row_offs = np.arange(0, first.height, window_size)
        window_tuples = [x + (window_size, window_size) for x in list(
            itertools.product(col_offs, row_offs))]
        winlist = [rasterio.windows.Window(*x) for x in window_tuples]
        windows = list(zip(range(len(winlist)), winlist))
        meta_proba = meta.copy()
        meta_proba.update(dtype=rasterio.dtypes.float32, nodata=-9999)
    bandcount = []
    for raster in rasters:
        with rasterio.open(raster, 'r') as src_r:
            bandcount.append(src_r.count)
    return windows, {'f_meta': f_meta, 'dst_meta': meta,
                     'dst_proba_meta': meta_proba,
                     'bandcount': np.sum(bandcount)}


def open_single_raster(raster_path, window=None):
    """Opens single raster and returns numpy array.

    Array shape: nbands x nrows x ncols.

    Args:
        raster_path(str): location of the raster file
        window (rasterio.Window): subset window

    returns:
        ndarray:    Raster data
        meta (dict): dictionary with raster metadata
    """

    with rasterio.open(raster_path, 'r') as raster:
        raster_array = raster.read(window=window)
        meta = raster.meta
        meta.update(
            width=raster_array.shape[2],
            height=raster_array.shape[1],
            transform=raster.window_transform(window))
        return raster_array, meta


def get_raster_date(raster):
    """Gets the date of a tif file from the name. The pattern that is expected
    is YYYY-MM-DD.

    If there are multiple dates with that format in the filename, only the
    first one will be used

        Args:
            raster (str) : Path to raster file
        returns:
            file_date (str): date of the raster in string format
    """
    file_date = re.findall(r'\d\d\d\d-\d\d-\d\d', raster)[0]
    return file_date

def raster_warp_dst(raster):
    """Get the warp paramters of a raster:

        Args:
            raster(str): path to a raster file

        Returns:
            warp_dst (dict): warp parameter dict
    """
    with rasterio.open(raster) as first:
        warp_dst = {
            "width": first.width,
            "height": first.height,
            "transform": first.transform,
            "crs": first.crs
        }
    return warp_dst

def count_bands(raster):
    """Returns number of bands of a raster

        Arg:
            raster (str): path to raster file

    """
    with rasterio.open(raster, 'r') as src:
        return src.count

def verify_and_count_bands(rasters):
    """Counts the number of bands for all the rasters and makes sure they all
    have the same number bands. Then it returns the number of bands that the
    rasters have

        Args:
            rasters(list): list of raster files

        Returns:
            b_count (int): number of bands in each raster

    """
    n_bands_list = [count_bands(raster) for raster in rasters]
    count = set(n_bands_list)
    if len(count) > 1:
        UTILS_RASTER_LOGGER.error("Number of bands per raster is not equal. "
                                  "Quitting...")
        sys.exit()
    return count.pop()



"""General utilities"""
import contextlib
import inspect
import json
import logging
import os
import pickle
import shutil
import sys
import tempfile
from typing import List, Callable
from uuid import uuid4
from zipfile import ZipFile

import joblib
from joblib import Parallel, delayed
import numpy as np
import rasterio
from sklearn.impute import SimpleImputer
from tqdm import tqdm

from classifier.settings import WORKSPACE, PARAMETERS, RASTER_EXTENSIONS

UTILS_GENERAL_LOGGER = logging.getLogger(__name__)

def get_available_model_args(model_args, function):
    """Gets the available arguments for the specific function
        Args:
            model_args(dict) : arguments to check
            function(func): function to check the kwargs from

        Returns:
            model_algorithm_args(list): model arguments that belong to function

    """
    kwarglist = inspect.getfullargspec(function)[0]
    model_algorithm_args = [x for x in model_args.keys() if x in kwarglist]
    return {k: model_args[k] for k in model_algorithm_args}


def progress(count, total, status=''):
    """A simple progress bar for the command line

    Args:
        count (int) the count between 0 and total
        total (int) the last iteration

    """
    count += 1
    bar_len = 60
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    p_bar = '=' * filled_len + '-' * (bar_len - filled_len)
    sys.stdout.write(f'[{p_bar}] {percents}% ...{status}\r')
    sys.stdout.flush()



def cli_init(output_name, rasters, overwrite, config_location, rois= None):
    """Initialize everything before starting real work.

    Args:
        output_name (str): name of the output directory in workspace
        raster (str): raster path(s) for input
        overwrite (bool): overwrite existing folder
        config_location (str): location of the config file
        rois (str): rois path for input


    Returns:
        Output_directory: path to output directory
        Rasters (list): list of separate raster files
        rois (str): path of rois workspace appended
        Args (dict): Dictionary of parameters to use

    """
    # Set Random Seed for reproducible results
    np.random.seed(0)

    rasters = get_raster_paths(WORKSPACE, rasters)
    config_dict = read_config(config_location)

    output_directory = create_output_dir(WORKSPACE,
                                         output_name,
                                         overwrite)

    config_dict['tmp_dir'] = os.path.join(output_directory, 'tmp')
    os.mkdir(config_dict['tmp_dir'])
    init_logger(output_directory, config_dict['app_log_level'])
    params_str = '\n'.join(
        [
            f'{key} :  {value}' for key, value in sorted(config_dict.items())
        ]
    )
    UTILS_GENERAL_LOGGER.info(
        "\nRunning the Classifier with the following parameters:"
        "\n  %s", params_str)
    if rois is not None:
        rois = os.path.join(WORKSPACE, rois)
    return output_directory, rasters, rois, config_dict



def create_output_dir(workspace, name, overwrite):
    """Create output directory.

    Args:
        workspace: The workspace Path
        name: Location of the directory
        overwrite (bool): overwrite existing folder


    Returns:
        Output directory: Path of output directory

    """
    if name is None:
        name = str(uuid4())[:6]
        UTILS_GENERAL_LOGGER.info("No name argument found. Making new folder "
                                  "called: %s", name)
    output_directory = os.path.join(workspace, name)
    if os.path.exists(output_directory):
        if overwrite:
            UTILS_GENERAL_LOGGER.warning("Overwriting existing directory!")
            shutil.rmtree(output_directory)
            os.mkdir(output_directory)
        else:
            UTILS_GENERAL_LOGGER.error(
                "Directory with name %s already exists. Either "
                "leave --name out,remove the directory %s, "
                "provide a unique name or turn on --overwrite.", name, name)
            sys.exit()
    else:
        os.mkdir(output_directory)
    return output_directory


def save_dict_as_json(out_file=None, dict_to_save=None):
    """Save default config values.

    Save default values for classifier parameters in json in current workspace.

    Args:
        out_file (str): file to write
        dict_to_save (dict): Dictionary to save

    """
    if out_file is None:
        out_file = os.path.join(WORKSPACE, 'config.json')
        dict_to_save = PARAMETERS
    with open(out_file, 'w', encoding='UTF-8') as config:
        config.write(json.dumps(dict_to_save,
                                sort_keys=True,
                                indent=1))


def read_config(config_file):
    """Read config parameter file and change defaults where necessary.

    Args:
        config_file (str): location of the config file
    returns:
        PARAMETERS (dict): dictionary with parameters
    """
    config_file = os.path.join(WORKSPACE, config_file)
    if os.path.isfile(config_file):
        with open(config_file, encoding='UTF-8') as json_config:
            config_dict = json.load(json_config)
        for key in config_dict.keys():
            PARAMETERS[key] = config_dict[key]
    else:
        UTILS_GENERAL_LOGGER.warning(
            "No local config file found. Using defaults")
    return PARAMETERS


def init_logger(output_directory, log_level):
    """Set and initialize all logging info.

    Args:
        output_directory: The location of the log file

    """
    logging.captureWarnings(True)
    logging.basicConfig(level=logging.DEBUG,
                        filename=os.path.join(output_directory, 'stdout.log'))
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)

    # # create console handler and set level to info
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setLevel(log_level)
    # # create formatter
    formatter = logging.Formatter(
        '%(asctime)s - %(name)s - %(levelname)s - %(message)s')
    # # add formatter to ch
    console_handler.setFormatter(formatter)
    # # add Handlers to logger
    logging.getLogger('').addHandler(console_handler)


def impute_values(dataset, args):
    """Impute values

    Uses the sklearn SimpleImputer to impute missing values.

    Args:
        dataset (DataFrame): DataFrame containing Nans
        args (dict): Parameters to use for imputation

    returns:
        dataset (DataFrame): DataFrame with Nans imputed
    """
    # remove inf if present
    if isinstance(dataset, np.ndarray):
        dataset = np.where(np.isinf(dataset), np.nan, dataset)
    else:
        dataset = dataset.replace([np.inf, -np.inf], np.nan)
    verbosity_level = 0 if args['app_log_level'] == 'INFO' else 2
    if np.isnan(dataset).all() and not args['app_imputation_strategy'] == \
                                       'constant':
        UTILS_GENERAL_LOGGER.warning(
            "Empty window found, Falling back to constant "
            "strategy")
        strategy = 'constant'
    else:
        strategy = args['app_imputation_strategy']

    imputer = SimpleImputer(strategy=strategy,
                            fill_value=args['app_imputation_constant'],
                            verbose=verbosity_level)

    return imputer.fit_transform(dataset)

def write_tifs(temp_dir, window, meta, result):
    """
    Writes the tifs of the individual windows
    Args:
        temp_dir: Temp dir location (str)
        window: rasterio window (tuple)
        meta: rasterio meta dictionary for output
        result: Result array

    Returns:
        Nothing

    """
    # Check output dir existence
    os.makedirs(temp_dir, exist_ok=True)
    meta['compress'] = 'deflate'
    with rasterio.open(
            os.path.join(
                temp_dir,
                f"c{window.col_off}_{window.row_off}.tif"
            ),
            'w',
            **meta
    ) as dst:
        dst.write_band(1, result)


def impute(data_array, config_dict):
    """Impute missing values for prediction

    Args:
        data_array (Numpy Array): Array containing missing values
        config_dict (dict): dictionary with all application parameters

    returns:
        Imputed data_array

    """
    shape = data_array.shape
    reshaped_array = data_array.reshape(shape[0], (shape[1] * shape[2])).T
    imputed_array = impute_values(reshaped_array, config_dict)
    imp_shape = imputed_array.shape
    if not shape[0] * shape[1] * shape[2] == imp_shape[0] * imp_shape[1]:
        UTILS_GENERAL_LOGGER.warning("Not all columns could be calculated. "
                                     "Falling  back to constant strategy for "
                                     "this chunk")
        config_dict['app_imputation_strategy'] = 'constant'
        imputed_array = impute_values(reshaped_array, config_dict)
    imputed_array = imputed_array.T.reshape(shape[0], shape[1], shape[2])
    return imputed_array


def save_model(model_dict, out_dir, config_dict):
    """Save the model as a pickle  and metadata  as a json and zip

        Args:
            model_dict (dict):  Contains the model as well as metadata
            out_dir (str):      Path where to write model file and meta
            config_dict (dict): Configuration parameters

    """
    to_save = model_dict.copy()
    meta_tmp = tempfile.mktemp()
    pickle_tmp = tempfile.mktemp()

    # Take the model out of the dictionairy
    model = to_save.pop('model')

    # Save the model as pickle
    with open(pickle_tmp, 'wb') as model_file:
        pickle.dump(model, model_file)

    # Save meta as json
    save_dict_as_json(meta_tmp, to_save)
    to_write = {
        meta_tmp: f"{config_dict['name']}_meta.json",
        pickle_tmp: f"{config_dict['name']}.model"
    }

   # Zip to output directory
    zipfile_path = os.path.join(out_dir, 'model.zip')
    write_zipfile(to_write, zipfile_path)


def write_zipfile(files_to_write, zipfile_path):
    """Writes files to a zipfile

    Args:
        files_to_write (dict): dictionary of path, name of the files to
                               write to the zipfile (e.g.
                               {
                                  '/workspace/weirdname.json':
                                  'nice_name.json'
                               }
        zipfile_path (str): Path to where the zipfile goes
    """
    with ZipFile(zipfile_path, 'w') as model_zip:
        for files in files_to_write:
            print(files, files_to_write[files])
            model_zip.write(files, files_to_write[files])

def get_raster_paths(workspace, rasters):
    """Get full paths for all rasters.

    Args:
        workspace: The workspace Path
        rasters: list of rasters from the cli

    Returns:
        list: rasters with their full paths

    """
    paths = []

    for raster in rasters:
        path = os.path.join(workspace, raster)
        if os.path.isdir(path):
            paths += [
                os.path.join(path, r)
                for r
                in sorted(os.listdir(path))
            ]
        else:
            paths.append(path)
    return [x for x in paths if os.path.splitext(x)[-1] in RASTER_EXTENSIONS]


def unzip_model_file(model_file):
    """Unzips a model file and returns paths to metadata and model pickle

    Args:
        model_file (str): path to the model zipfile

    Returns:
        meta_file (str): path to tmp metafile
        pickle_file (str): path to tmp pickle file
    """
    tmpdir = tempfile.mkdtemp()
    with ZipFile(model_file, 'r') as zipf:
        zipf.extractall(tmpdir)

    model_files = os.listdir(tmpdir)
    meta_file = os.path.join(
        tmpdir,
        [x for x in model_files if x.endswith('.json')][0]
    )
    pickle_file = os.path.join(
        tmpdir,
        [x for x in model_files if x.endswith('.model')][0]
    )
    return meta_file, pickle_file


def read_model(model_file):
    """Function to read the model file and return a pickle

    Args:
        path: The path to the model zipfile

    Returns:
        a dictionary containing a model, and metadata

    """
    # Unzip
    meta_file, pickle_file = unzip_model_file(model_file)
    # Open the metadata and print the modeltypes
    with open(meta_file, encoding='UTF-8') as json_config:
        model_dict = json.load(json_config)

    params_str = '\n'.join([f'{key} :  {value}'for key, value in
                            sorted(model_dict.items())])

    with open(pickle_file, 'rb') as pfile:
        model_dict['model'] = pickle.load(pfile)

    UTILS_GENERAL_LOGGER.info(
        "\nLoaded a saved model, with the following metadata:"
        "\n  %s", params_str)

    return model_dict

def parallel_function(func: Callable, iterable: List[dict],
                      ncpus: int = 1) -> list:
    """Runs a function in parallel using Joblib
        Args:
            func: Function to run
            iterable (list): List of kwargs for the function
            ncpus (int): number of cpus to use
        Returns:
            list of results
    """
    with tqdm_joblib(
            tqdm(total=len(iterable))
    ) as _:
        result = Parallel(
            n_jobs=ncpus)(delayed(func)(**x) for x in iterable)

    return result


@contextlib.contextmanager
def tqdm_joblib(tqdm_object):
    """Patch joblib to report a tqdm progress bar for many thread jobs

    Args:
        tqdm_object (tqdm.tqdm): Iterable decorated with progress bar

    Yields:
        tqdm.tqdm: Iterable decorated with progress bar
    """

    # pylint: disable=too-few-public-methods
    class TqdmBatchCompletionCallback(joblib.parallel.BatchCompletionCallBack):
        """ Class to pass completion callback to tqdm"""

        def __call__(self, out):
            """Updates tqdm and returns joblib call

            Args:
                 out (object): Any object

            Returns:
                object: Any object
            """
            tqdm_object.update(n=self.batch_size)
            return super().__call__(out)

    old_batch_callback = joblib.parallel.BatchCompletionCallBack
    joblib.parallel.BatchCompletionCallBack = TqdmBatchCompletionCallback
    try:
        yield tqdm_object
    finally:
        joblib.parallel.BatchCompletionCallBack = old_batch_callback
        tqdm_object.close()



"""File containing all model classes necessary to run the classifier"""
import logging
import os

import matplotlib.pyplot as plt
import pandas as pd
from sklearn.ensemble import IsolationForest
from sklearn.ensemble import RandomForestClassifier
from sklearn.model_selection import RandomizedSearchCV
from xgboost import XGBClassifier


# initiate logger
MODELS_LOGGER = logging.getLogger(__name__)


class BaseMixin:
    """A mixin class for all models. This allows for each algorithm to use the
    optimise and plotting functions"""

    def plot_time_vs_accuracy(self, out_dir):
        """Plotting the training time vs the accuracy.

        Args:
            time (array): The time values to plot
            accuracy(array) The accuracy values to plot
            out_dir: THe output directory (path)

        Returns:
            Nothing
        """
        time = self.cv_df['mean_fit_time']
        accuracy = self.cv_df['mean_test_score']
        fig = plt.figure()
        axis = fig.add_subplot(111)
        axis.scatter(time, accuracy)
        axis.set_xlabel("Time (s)")
        axis.set_ylabel("Accuracy (-) ")
        plt.tight_layout()
        plt.savefig(os.path.join(out_dir, 'Optimisation_time.png'), dpi=300)

    def random_optimise(self, trainx, trainy, out_dir,
                        optimise_iters=10):
        """Optimization of a parameter space using random samples from the
        parameter space

        Args:
            trainx (array) :   Training dataset features (X-values)
            trainy (array) :   Training dataset outputs (class names, y values)
            out_dir (string):  The output directory path
            optimise_iters (int)  The number of iterations for the optimisation.

        Returns:
            The best performing parameter combination  (dict)

        """
        MODELS_LOGGER.info("\n####-----Optimisation----#####\n")
        MODELS_LOGGER.info("Starting Optimisation. This might take a while....")
        clf = RandomizedSearchCV(self,
                                 self.parameter_matrix,
                                 verbose=0,
                                 refit=True,
                                 cv=3,
                                 return_train_score=True,
                                 n_jobs=-1,
                                 n_iter=optimise_iters
                                )
        clf.fit(trainx, trainy)
        MODELS_LOGGER.info("Best estimator: \n %s", clf.best_estimator_)
        self.cv_df = pd.DataFrame(clf.cv_results_)
        MODELS_LOGGER.debug("--Optimisation Results--\n %s",
                            self.cv_df[['mean_train_score', 'mean_test_score',
                                        'mean_fit_time']])
        parameter_set = self.cv_df['params'][self.cv_df['rank_test_score'] ==
                                             1].values[0]
        MODELS_LOGGER.debug("\nThe best parameter combination is:\n %s",
                            parameter_set)
        self.plot_time_vs_accuracy(out_dir)
        return parameter_set

# pylint: disable=too-many-ancestors
class RandomForest(RandomForestClassifier, BaseMixin):
    """RandomForest class; child of RandomForestClassifier from sk-learn"""

    def __init__(self,
                 criterion="gini",
                 max_depth=None,
                 min_samples_split=2,
                 min_samples_leaf=1,
                 min_weight_fraction_leaf=0.,
                 max_features="auto",
                 max_leaf_nodes=None,
                 min_impurity_decrease=0.,
                 min_impurity_split=None):
        super().__init__()

        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.min_impurity_split = min_impurity_split

        self.parameter_matrix = {'max_features': ['auto', 'sqrt', 'log2'],
                                 'max_depth': [None, 1, 3, 10, 20000]}


class XGBoost(XGBClassifier, BaseMixin):
    """XGBoost class, child of XGBoostclassifier model from sk-learn"""

    def __init__(self):
        super().__init__()
        self.parameter_matrix = {'learning_rate': [0.1, 0.2, 0.3],
                                 "n_estimators": [10, 20, 50, 100, 200],
                                 'max_depth': [1, 3, 5, 10]
                                }


class SingleClass(IsolationForest, BaseMixin):

    """Singleclass class, child of IsolationForest model from sk-learn"""


    def __init__(self):
        super().__init__()

        self.parameter_matrix = {"n_estimators": [10, 20, 50, 100, 200],
                                 'max_samples': ['auto', 10, 50, 100, 200],
                                 "max_features": [0.1, 0.3, 0.5, 1.0]
                                }



"""Settings file for Classifier"""
from sklearn.cluster import KMeans, MiniBatchKMeans
from classifier.models import RandomForest, XGBoost, SingleClass

WORKSPACE = '/workspace/'

RASTER_EXTENSIONS = frozenset(['.tif', '.vrt', '.jp2'])

US_ALGORITHMS = ['us_kmeans', 'us_kmeans_minibatch']

# ##----THE ALGORITHMS---###
ALGORITHMS = ["randomforest",
              "xgboost",
              "singleclass",
              "us_kmeans",
              "us_kmeans_minibatch"
             ]
CLASSIFIERS = [
    RandomForest,
    XGBoost,
    SingleClass,
    KMeans,
    MiniBatchKMeans
    ]
ALGORITHM_DICT = dict(zip(ALGORITHMS, CLASSIFIERS))

PARAMETERS = {
    'app_algorithm': 'randomforest',
    # randomforest, xgboost, singleclass,us_kmeans
    'app_window': 1024,  # Any int, preferably within 2^x
    'app_model': None, # Model File location
    'app_samples': None,  # Samples file location (csv)
    'app_log_level': 'INFO',  # Logging level
    'app_threads': -1,  # #of threads
    'app_imputation': False,  # Use simple imputation for missing values
    'app_imputation_strategy': 'mean',  # Strategy for imputation. mean,
    'app_rasters_are_timeseries': False,  # Whether or not the input rasters
    # are TS
    # median, most_frequent, constant or randomforest
    'app_imputation_constant': -9999,  # imputation constant
    'su_probability': False,  # Output Probability Map
    'su_all_probabilities': False,
    'su_optimize': False,  # Optimize the model parameters
    'su_optimize_number': 10,  # Number of iterations for optimization
    'su_search_max_features': ['auto', 'sqrt', 'log2'],
    'su_search_max_leaf_nodes': [3, 5, 7],
    'su_search_max_depth': [None, 1, 3, 10, 20000],
    'su_search_n_estimators': [10, 50, 100],
    'su_single_class_treshold': 0,  # Treshold to use for single class
    'su_boxplots': False,  # Plot Boxplots for samples
    'su_remove_outliers': True,  # Remove outliers from the training data
    'su_rf_tree_depth': None,  # Depth of the trees in RF model.
    'us_nclasses': 2,  # Number of classes for unsupervised
    'us_trainfraction': 1.0,  # Fraction of raster used for training
    'acc_perform_assesment': True,  # Perform accuracy assessment
    'acc_testfraction': 0.25,  # Fraction of data to use for training
    'model_save': False,  # Save a model file which can be re-used
}



"""An sk-learn classifier for landcover classification"""

import logging
import os
import time
import shutil
import sys

import fiona
import pandas as pd

from classifier.dataprep import createdataset, outlier_removal
from classifier.train import train_dataset
from classifier.accuracy import write_confusion_matrix, plot_feature_importances
from classifier.predict import prediction
from classifier.unsupervised import make_input_array, train_kmeans
from classifier.utils.general import save_model, read_model, parallel_function
from classifier.utils.raster import stitch, get_meta
from classifier.timeseries import get_timeseries_samples
from classifier.settings import WORKSPACE, US_ALGORITHMS

START = time.time()
CLASSIFIER_LOGGER = logging.getLogger(__name__)


def read_samples(samples, remove_outliers=True):
    """Read a samples file.

    Args:
        samples(path): CSV file containing samples
        remove_outliers (Bool): Flag for removing outliers from samples

    Returns:
        sample_df(DataFrame): DF containing samples

    """
    sample_df = pd.read_csv(samples, index_col=0)
    if remove_outliers:
        sample_df = outlier_removal(sample_df)
    return sample_df

def get_rois_extent(rois):
    """

    Args:
        rois (str): path to a rois file

    Returns:
        bounds of rois (list): ulx, uly, llx, lly coordinates

    """
    with fiona.open(rois, "r") as shapefile:
        return shapefile.bounds

def gather_samples(rois, rasters, out_dir, config_dict):
    """Gather samples from a rois file and raster files

    Args:
        rois: The path to a rois file
        rasters: A list of raster files
        out_dir: Path where to write the dataset file
        config_dict: The complete classifier configuration dictionary

    Returns:
        A dataframe containing pixel values and classes

    """
    # Train a model and continue
    samples = createdataset(
        rasters,
        os.path.join(WORKSPACE, rois),
        out_dir,
        config_dict)
    samples.to_csv(os.path.join(out_dir, 'samples.csv'))
    return samples


def train(samples, rasters, out_dir, config_dict, rois_extent):
    """Train a model using samples and rasters

    Args:
        samples: Samples dataframe containing pixel values and class values
        rasters: A list of raster files
        out_dir: Path where to write the dataset file
        config_dict: The complete classifier configuration dictionary
        rois_extent: The extent of the training data (if any, else None)

    Returns:
        The return value. True for success, False otherwise.

    """
    CLASSIFIER_LOGGER.info("\n####-----Training----#####\n")
    dataset = samples
    windows, _ = get_meta(rasters, config_dict['app_window'])
    if config_dict['app_algorithm'] in US_ALGORITHMS and len(samples) < 1:
        # Do unsupervised
        array = make_input_array(
            rasters,
            windows,
            config_dict
        )
        model_dict = train_kmeans(array,
                                  config_dict)
    else:  # All supervised methods
        model_dict, test = train_dataset(
            dataset,
            config_dict['app_algorithm'],
            out_dir,
            config_dict
            )
        if config_dict['app_algorithm'] in ['randomforest', 'xgboost'] \
                and config_dict['acc_perform_assesment']:
            # Do the accuracy analysis
            cm_fn = os.path.join(out_dir, 'confusion_matrix')
            fi_fn = os.path.join(out_dir, 'feature_importance.png')
            write_confusion_matrix(model_dict, test, cm_fn)
            plot_feature_importances(model_dict, fi_fn)
    # save the model as a python pickle
    if not rois_extent is None:
        model_dict['rois_bounds'] = rois_extent
    if config_dict['model_save']:
        save_model(model_dict, out_dir, config_dict)
    CLASSIFIER_LOGGER.info("\nFinished Training\n")
    return model_dict


def predict(model_dict, rasters, out_dir, config_dict):
    """Prediction function using a trained model and raster files

    Args:
        model_dict: A trained model
        rasters: A list of raster files
        out_dir: Path where to write the dataset file
        config_dict:  The command line arguments dictionary

    """
    threads = config_dict['app_threads']
    CLASSIFIER_LOGGER.info("\n####-----Prediction----#####\n")
    windows, meta = get_meta(rasters, config_dict['app_window'])

    iterable = [
        {
            'window': window[1],
            'rasters': rasters,
            'model_dict': model_dict,
            'meta': meta,
            'config_dict': config_dict
        }
        for window in windows
    ]
    if threads > 1 or threads == -1:

        parallel_function(prediction,
                              iterable,
                              threads)
    else:
        for wins in iterable:
            prediction(**wins)

    # ##--------------------STITCHING-----------------------------------###

    CLASSIFIER_LOGGER.info("\n####-----Stitching----#####\n")

    # Run the gdalwarp command in the command line
    stitch(out_dir, meta, config_dict)
    CLASSIFIER_LOGGER.info("Cleaning up..")
    shutil.rmtree(config_dict['tmp_dir'])
    CLASSIFIER_LOGGER.info(
        "Total run time was %i seconds", (int(time.time() - START)))


def train_and_predict_with_samples(samples, rasters, out_dir, config_dict,
                                   rois_extent=None):
    """Train model using samples

    Args:
        samples: Samples dataframe containing pixel values and class values
        rasters: A list of raster files
        out_dir: Path where to write the dataset file
        config_dict:  The command line arguments dictionary

    Returns:
        The return value. True for success, False otherwise.

    """
    model_dict = train(samples, rasters, out_dir, config_dict, rois_extent)
    predict(model_dict, rasters, out_dir, config_dict)


def gather_train_predict_rois(rois, rasters, out_dir, config_dict):
    """Gather samples from rois file and rasters and continue with training and
    prediction

    Args:
        rois: The path to a rois file
        rasters: A list of raster files
        out_dir: Path where to write the dataset file
        config_dict:  The command line arguments dictionary

    Returns:
        The return value. True for success, False otherwise.

    """
    samples = gather_samples(rois, rasters, out_dir, config_dict)
    rois_extent = get_rois_extent(rois)
    return train_and_predict_with_samples(samples,
                                          rasters,
                                          out_dir,
                                          config_dict,
                                          rois_extent)

def classify_(rasters, config_dict, out_dir, rois=None):
    """Entry point function for pixel-based classification

        Args:
            rasters (list): List of raster paths
            config_dict (dict): CLI configuration dict
            out_dir (str): Path to output directory
            rois (str): Path to rois file

    """
    if not config_dict['app_model'] is None:
        model = read_model(config_dict['app_model'])
        predict(model, rasters, out_dir, config_dict)
    elif not config_dict['app_samples'] is None:
        samples = read_samples(
            config_dict['app_samples'],
            config_dict['su_remove_outliers'])

        train_and_predict_with_samples(samples,
                                       rasters,
                                       out_dir,
                                       config_dict)

    elif rois is not None:
        gather_train_predict_rois(
            rois,
            rasters,
            out_dir,
            config_dict)
    else:
        CLASSIFIER_LOGGER.info("No models, samples or rois provided. Doing "
                               "unsupervised classification")
        if not config_dict['app_algorithm'] in US_ALGORITHMS:
            config_dict['app_algorithm'] = 'us_kmeans'
        train_and_predict_with_samples([],
                                       rasters,
                                       out_dir,
                                       config_dict)
def classify_timeseries(rasters: list,
                        config_dict: dict,
                        out_dir: str,
                        rois: str):
    """Main function to classifiy timeseries.

    For now, only from start to finish is supported. ie, supplying a model
    or samples does not work.

        Args:
            rasters (list): List of raster paths
            config_dict (dict): CLI configuration dict
            out_dir (str): Path to output directory
            rois (str): Path to rois file

    """
    # Gather samples and do imputation.
    CLASSIFIER_LOGGER.info("Now Getting Samples and doing imputation")
    samples_df = get_timeseries_samples(rasters, rois, out_dir, config_dict)
    # Reshape it
    samples = samples_df.T
    samples['class'] = samples.index
    if config_dict['app_algorithm'] == 'unsupervised':
        CLASSIFIER_LOGGER.error("Unsupervised classification not supported yet")
        sys.exit()
    model = train(samples,
                  rasters=rasters,
                  out_dir=out_dir,
                  config_dict=config_dict,
                  rois_extent=None)

    predict(model, rasters, out_dir, config_dict)

if __name__ == "__main__":
    pass



""" classifier metadata """
import os
import pkg_resources

try:
    __version__ = pkg_resources.get_distribution("s11-classifier").version
except pkg_resources.DistributionNotFound:
    __version__ = os.environ['VERSION']



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Jan 15 08:58:19 2018

@author: rensmasselink
"""
import logging
from multiprocessing import Manager
import numpy as np
import rasterio
from rasterio.enums import Resampling
from rasterio.vrt import WarpedVRT

from classifier.dataprep import check_training_data_nodata
from classifier.train import ALGORITHM_DICT
from classifier.utils.general import parallel_function
UNSUPERVISED_LOGGER = logging.getLogger(__name__)

def make_input_array(rasters, windows, config_dict):
    """ Create an input array for unsupervised classification

    Function for creating the input array for the training of the
    unsupervised models. A train_size relative to the entire dataset can be
    set to reduce the amount of memory needed

       Args:
           rasters (list)      input rasters
           windows (list)      A list of rasterio Windows to use
           config_dict (dict)   The parameters for making an input array

        Returns:
            A data array to use in model training"""
    # Create a 2d array which can be passed into the classifier
    bandcount = []
    for raster in rasters:
        with rasterio.open(raster, 'r') as src_r:
            bandcount.append(src_r.count)
    # Random index for choosing the windows to use
    subset_windows = np.random.choice(
        np.array(windows)[:, 1],
        size=int(len(windows) * config_dict['us_trainfraction']),
        replace=False
    )
    n_windows = len(subset_windows)
    assert n_windows > 0, (
        "Not enough subset data. Please increase -us_train_size parameter"
    )

    UNSUPERVISED_LOGGER.info(
        "\nUsing %i windows for training",
        n_windows
    )
    with Manager() as manager:
        mp_list = manager.list()
        kwargs = [
            {
                'mp_list': mp_list,
                'rasters': rasters,
                'window': window
            }
            for window in subset_windows
        ]
        parallel_function(
            get_window_data_for_us_train,
            kwargs,
            ncpus=config_dict['app_threads'])
        data_array = np.concatenate(mp_list)

    #Check for missing values and impute if necessary
    data_array = check_training_data_nodata(data_array, config_dict)

    return data_array[~np.isnan(data_array).any(axis=1)]

def get_window_data_for_us_train(mp_list, rasters, window):
    """Gets the pixel values for window and adds it to an MP list

        Args:
            mp_list (mp.list): list to append the values to. List is changed
                               inplace and is not returned
            rasters (list): List of rasters to get data from
            window (rasterio.Window): Window to process

    """
    data_array = []
    with rasterio.open(rasters[0], 'r') as template:
        # Create a vrt with the window and add the data to the array
        for raster in rasters:
            with rasterio.open(raster, 'r') as src:
                with WarpedVRT(
                    src,
                    resampling=Resampling.bilinear,
                    meta=template.meta,
                    width=template.meta['width'],
                    height=template.meta['height'],
                    crs=template.crs,
                    transform=template.transform
                ) as vrt:
                    data_array.append(vrt.read(window=window))
    old_shape = np.shape(data_array)
    mp_list.append(
        np.reshape(
            data_array,
            (old_shape[0]*old_shape[1], old_shape[2]*old_shape[3])
        ).transpose()
    )

def train_kmeans(train_array, config_dict):
    """Train the kmeans model

        Args:
            train_array (array) A data array with columns (bands) and rows (
            pixels)
            config_dict (dict): Configuration Dictionary

        Returns:
            A dictionary conaining the algorithm name, the trained model and an
            empty label key to match the dicts from the supervised
            classifications
    """
    algorithm = ALGORITHM_DICT[config_dict['app_algorithm']]
    n_classes = config_dict['us_nclasses']

    UNSUPERVISED_LOGGER.info("Now Training Model. This might take some time...")

    kmeans = algorithm(n_clusters=n_classes)
    kmeans.fit(train_array)
    return {'app_algorithm': config_dict['app_algorithm'],
            'model': kmeans,
            'labels': None}



"Training module for classifier"
import numpy as np

from sklearn.model_selection import train_test_split

from classifier.utils.general import get_available_model_args
from classifier import __version__ as classifier_version
from classifier.settings import ALGORITHM_DICT


def train_dataset(dataset, algorithm, out_dir,
                  config_dict):
    """Train the model using a dataset

    Args:
        dataset         Dataset containing features in the columns wiht one
                        column named "class" which contains class labels or
                        \numbers
        algorithm       The algorithm name
        out_dir         Path where to write dataset file
        config_dict     command line arguments dictionary

    Returns:
        model_dict      A dictionary containing the name, model and label
                        encoder.
        test            A test dataset which was not used during training
    """
    # Encode the labels
    labels = np.unique(dataset['class'].values).tolist()
    labels = dict(zip(range(len(labels)), labels))
    # Split the dataset,
    model, xcols, test = init_model_and_train(dataset,
                                              algorithm,
                                              out_dir,
                                              config_dict)

    model_dict = {'app_algorithm': algorithm,
                  'model': model,
                  'labels': labels,
                  'names': xcols,
                  'version': classifier_version}
    return model_dict, test

def set_model_parameters(algorithm, algorithm_args, y_train):
    """
    Set the model parameters
    Args:
        algorithm:  Algorithm name
        algorithm_args: Algorithm arguments
        y_train: The training dataset classes

    Returns:
        parametrized model

    """
    model_type = ALGORITHM_DICT[algorithm]
    model = model_type()
    model_algorithm_args = get_available_model_args(algorithm_args, model_type)
    model.set_params(**model_algorithm_args)
    n_class = len(set(y_train))
    if algorithm == 'xgboost' and n_class < 3:
        model.set_params(**{'objective': 'binary:logistic'})
    return model

def init_model_and_train(dataset, algorithm, out_dir,
                         config_dict):
    """
    Set the model parameters and train it
    Args:
        dataset (Array) : The dataset for input in the model (array)
        algorithm: The algorithm to use
        out_dir: The output directory

    Returns:
        Trained model
        Names of bands
        test dataset

    """
    optimize = config_dict['su_optimize']
    optimize_iters = config_dict['su_optimize_number']
    test_size = config_dict['acc_testfraction']
    train, test = train_test_split(dataset, test_size=test_size)
    xcols = [x for x in train.columns if 'class' not in x]
    x_train = train[xcols].values
    y_train = np.ravel(train[[x for x in train.columns if 'class' in x]].values)
    # Get the model and fit
    model = set_model_parameters(algorithm, {}, y_train)
    if optimize:
        try:
            model.parameter_matrix = {
                'max_features': config_dict['su_search_max_features'],
                'max_depth': config_dict['su_search_max_depth'],
                'max_leaf_nodes': config_dict['su_search_max_leaf_nodes'],
                'n_estimators': config_dict['su_search_n_estimators']}
            algorithm_args = model.random_optimise(
                x_train,
                y_train,
                out_dir,
                optimize_iters
                )
            model = set_model_parameters(algorithm, algorithm_args, y_train)
        except AttributeError as no_attribute:  # if there is no param. matrix
            raise AttributeError from no_attribute
    # y_train_encoded = sample_labels_encoder.transform(y_train)
    model.fit(x_train, y_train)
    return model, xcols, test



"Accuracy assessment and plotting for classifier"
import itertools
import logging
import time

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
from sklearn.metrics import confusion_matrix, cohen_kappa_score
import fiona
import rasterio
from rasterio import mask
from shapely.geometry import shape
from shapely.ops import cascaded_union


ACCURACY_LOGGER = logging.getLogger(__name__)


def write_confusion_matrix(model_dict, test_set, cm_fn, plot=True, csv=True):
    """ Uses a sklearn confusion matrix (np 2d array) to write to a csv file
    that also contains overall metrics. Also plots the confusion matrix for
    easier viewing.

        Args:
            model_dict (dict):dictionary with the name, model and label encoder
            test_set (pd.DataFrame): test dataset not used during training
            cm_fn (str): The filename of the output csv
            plot (str): Whether or not to plot the figure
            csv (str): Whether or not to write csv file

        returns:
            nothing
    """
    # Convert dict and dataframe to arrays
    x_test = test_set[[x for x in test_set.columns if 'class' not in x]].values
    y_test = np.ravel(
        test_set[[x for x in test_set.columns if 'class' in x]].values)
    preds = model_dict['model'].predict(x_test)
    compute_confusion_matrix(y_test, preds, cm_fn, plot, csv)


def compute_confusion_matrix(y_test, preds, cm_fn, plot=True, csv=True):
    """ Uses a sklearn confusion matrix (np 2d array) to write to a csv file
    that also contains overall metrics. Also plots the confusion matrix for
    easier viewing.

        Args:
            y_test (np.array): Ground truth (correct) target values
            preds (np.array): Estimated targets as returned by a classifier
            cm_fn (str): The filename of the output
            plot (str): Whether or not to plot the figure
            csv (str):  Whether or not to write the csv
    """

    # Write and plot confusion Matrix
    cm_labels = sorted(list(set(np.unique(y_test)) | set(np.unique(
        preds))))
    kappa = cohen_kappa_score(y_test, preds, labels=cm_labels)
    conf_matrix = confusion_matrix(y_test, preds, labels=cm_labels)

    # Metrics: overall mean accuracy and kappa
    ACCURACY_LOGGER.info("\n####-----Accuracy Assessment-----#####\n")
    metrics_list = []
    metrics_list.append(
        f'Overall Accuracy {np.trace(conf_matrix)/np.nansum(conf_matrix)}'
    )
    metrics_list.append(f'Kappa {kappa:.4f}')
    for metric in metrics_list:
        ACCURACY_LOGGER.info(metric)

    if csv:
        df_cm = pd.DataFrame(
            data=conf_matrix, index=cm_labels, columns=cm_labels)
        # Total of the rows
        df_cm['Total'] = df_cm.sum(axis=1)
        # Accuracy
        df_cm['Accuracy'] = np.diag(df_cm[cm_labels]) / df_cm['Total']
        # Appending the numbers to the df
        df_cm = df_cm.append(df_cm.sum(axis=0).rename('Total')).append(
            (pd.Series(np.diag(conf_matrix) / np.sum(conf_matrix, axis=0),
                       index=cm_labels).rename('Reliability')))
        df_cm.to_csv(cm_fn+'.csv', sep=';', float_format='%.4f')

    if plot:
        # Plotting if necessary
        try:
            # Plot normalized confusion matrix
            plt.figure()
            plot_confusion_matrix(conf_matrix,
                                  classes=cm_labels,
                                  normalize=True
                                 )
            plt.savefig(cm_fn + '_plot.png', dpi=150)
        except AttributeError:  # Not all models have FIs, so skip
            pass


def plot_confusion_matrix(conf_matrix, classes,
                          normalize=False,
                          cmap=plt.cm.Blues,
                          labeldict=None):
    """
    This function prints and plots the confusion matrix.
    Normalization can be applied by setting `normalize=True`.
    """
    if normalize:
        conf_matrix = conf_matrix.astype('float') /\
                      conf_matrix.sum(axis=1)[:, np.newaxis]
    if labeldict is not None:
        classes = map(labeldict.get, classes)
    plt.imshow(conf_matrix, interpolation='nearest', cmap=cmap)
    plt.title("Confusion Matrix")
    plt.colorbar()
    tick_marks = np.arange(len(classes))
    plt.xticks(tick_marks, classes, rotation=90, fontsize=6)
    plt.yticks(tick_marks, classes, fontsize=8)

    fmt = '.2f' if normalize else 'd'
    thresh = conf_matrix.max() / 2.
    for i, j in itertools.product(range(conf_matrix.shape[0]),
                                  range(conf_matrix.shape[1])):
        plt.text(j, i, format(conf_matrix[i, j], fmt),
                 horizontalalignment="center",
                 color="white" if conf_matrix[i, j] > thresh else "black",
                 fontsize=4)

    plt.tight_layout()
    plt.ylabel('True label')
    plt.xlabel('Predicted label')


def plot_feature_importances(model_dict, outfile):
    """ Plot the feature importances of the forest"""
    importances = model_dict['model'].feature_importances_
    indices = np.argsort(importances)[::-1]
    labels = [model_dict['names'][x] for x in indices]
    nr_features = len(model_dict['names'])
    fig, axis = plt.subplots()
    axis.set_title("Feature importances (band numbers)")
    # Get STD whenever possible
    if model_dict['app_algorithm'] == 'randomforest':
        std = np.std([tree.feature_importances_ for tree in model_dict[
            'model'].estimators_], axis=0)
        axis.bar(range(nr_features), importances[indices],
                 color="r", yerr=std[indices], align="center")
    else:
        axis.bar(range(nr_features), importances[indices],
                 color="r", align="center")
    axis.set_xticks(range(nr_features))
    axis.set_xticklabels(labels, rotation=45, ha="right")
    axis.set_xlim([-1, nr_features])
    fig.tight_layout()
    plt.savefig(outfile, dpi=300)


def __gather_samples_for_roi(roi, rasterfile):
    """Get sample values from rasters

        Args:
            roi (fiona shape): region of interest
            raster (str): raster file name
            warp_dst (dict): warp specification

        Returns:
            np.array [nsamples, nfeatures] of samples from within rois.

    """
    with rasterio.open(rasterfile) as src:
        selection, _, window = mask.raster_geometry_mask(
            src,
            [roi["geometry"]],
            crop=True)
        raster_subset = src.read(window=window)

        # rasterio convention: outside shape=True, inside=False. We invert.
        samples_per_raster = raster_subset[:, ~selection]
        roi_values = np.asarray(samples_per_raster).flatten()

    return roi_values


def assess_accuracy(raster, rois, cm_fn, subset=None):
    """Accuracy assessment
        Args:
            raster (str): filename of classication raster
            rois (str): rois filename
            cm_fn (str): filename of output
            subset (str): filename of subset used for classication
    """

    start = time.time()

    # Is subset is given, exclude its polygons from accuracy assessment
    if subset:
        polygons = [
            shape(feature['geometry']) for feature in fiona.open(subset)]
        union_subset = cascaded_union(polygons)

    sample_labels = []
    all_samples = []
    counter = 0
    with fiona.open(rois, "r") as shapefile:
        for roi in shapefile:
            if subset and shape(roi['geometry']).intersects(union_subset):
                counter += 1
                continue
            roi_samples = __gather_samples_for_roi(
                roi,
                raster
            )
            if roi_samples is not None:
                all_samples += list(roi_samples.astype(np.float))
                roi_id = int(roi['properties']['id'])
                sample_labels += len(roi_samples) * [roi_id]
        all_samples = np.asarray(all_samples)
        sample_labels = np.asarray(sample_labels)

        ACCURACY_LOGGER.info(
            "Using %i groundtruth polygons, \
                excluding %i already used in classifier",
            len(shapefile),
            counter)

    #write confusion matrix
    compute_confusion_matrix(
        sample_labels, all_samples, cm_fn, plot=True, csv=False)

    ACCURACY_LOGGER.info(
        "Total run time was %i seconds", (int(time.time() - start)))


if __name__ == "__main__":
    START = time.time()
    RASTER = "classification.tif"
    ROIS = "merged_gewassen.geojson"
    SUBSET = "subset_merged_gewassen.geojson"
    CM_FN = "confusion_matrix_total"

    assess_accuracy(RASTER, ROIS, CM_FN, SUBSET)

    ACCURACY_LOGGER.info(
        "Total run time was %i seconds",
        (int(time.time() - START))
    )



# -*- coding: utf-8 -*-
# (c) Satelligence, see LICENSE.
# pylint: skip-file
from setuptools import setup
import setuptools
import os

version = '3.0.0'

long_description = open('README.md').read()

test_requirements = [
    'pytest'
]

setup(
    name='s11-classifier',
    version=version,
    description="Classifier",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Satelligence",
    author_email='team@satelligence.com',
    url='https://gitlab.com/satelligence/classifier',
    packages=setuptools.find_packages(),
    package_dir={
        'classifier': 'classifier'
    },
    include_package_data=True,
    install_requires=[
        'matplotlib>=3.0.3, <4.0.0',
        'numpy>=1.17.5, <2.0.0',
        'pylint>=2.12.0, <3.0.0',
        'pytest>=5.2.0, <6.0.0',
        'rasterio>=1.1.8, <2.0.0',
        'scikit_learn==1.0.2',
        'xgboost>=1.1.1, <2.0.0',
        'boto3>=1.16.63, <2.0.0',
        'folium>=0.12.1, <1.0.0',
        'geopandas>=0.8.2, <1.0.0',
        'geojson>=2.5.0, <3.0.0',
        'click>=7.1.2, <8.0.0',
        'rasterstats>=0.15.0, <1.0.0',
        'rtree>=0.9.7, <1.0.0',
        'xarray>=0.12.3, <1.0.0',
        'python-dateutil>=2.8.1, <3.0.0',
        'jinja2>=2.11.3, <3.0.0'

    ],
    license="Apache-2.0",
    zip_safe=False,
    python_requires='>=3.5'
)



"""All the necessary functions for preparing the data for the classification
model can be found here.."""
import base64
import logging
import os
import sys
import warnings

import fiona
import folium
from folium import IFrame
import geojson
import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import rasterio
from rasterio import mask
from rasterio.enums import Resampling
from rasterio.vrt import WarpedVRT
from shapely.geometry import shape, MultiPoint
from sklearn.ensemble import IsolationForest

from classifier.utils.general import progress, impute_values
from classifier.utils.raster import raster_warp_dst

# Set Logging levels on Rasterio and Fiona to not get too many logging msg
RASTERIO_LOGGER = logging.getLogger('rasterio')
RASTERIO_LOGGER.setLevel(logging.CRITICAL)
FIONA_LOGGER = logging.getLogger('fiona')
FIONA_LOGGER.setLevel(logging.CRITICAL)
DATAPREP_LOGGER = logging.getLogger(__name__)


def check_training_data_nodata(dataset, config_dict):
    """
    Check whether there is nodata in the training dataset
    Args:
        dataset (DataFrame or Numpy Array):Dataset with gathered sample data
        config_dict (dict): CLI configuration dictionary

    Returns:
        dataset without nans

    """
    #Check if nans are present
    contains_nans = False
    if isinstance(dataset, np.ndarray):
        if np.isnan(dataset).any():
            contains_nans = True
    else:
        if dataset.isnull().values.any():
            contains_nans = True

    if contains_nans and not config_dict['app_imputation']:
        DATAPREP_LOGGER.warning("""The sampled dataset contains nan's.
        Imputation is set to False so all rows containing NaNs will be 
        deleted. If you want to prevent this from happening, please set the 
        app_imputation parameter to True and choose a strategy and constant 
        if necessary.""")
        if isinstance(dataset, np.ndarray):
            dataset = dataset[~np.isnan(dataset).any(axis=1)]
        else:
            dataset = dataset.dropna(how='any')
    elif contains_nans:
        DATAPREP_LOGGER.info("Found missing values. Doing imputation...")
        if config_dict['app_imputation_strategy'] == 'randomforest':
            DATAPREP_LOGGER.error("Random Forest strategy only allowed for "
                                  "timeseries, please select another strategy")
            sys.exit()
        dataset = impute_values(dataset, config_dict)
    return dataset


def check_training_extent():
    """Check whether the extent of the rois falls outside of the extent of the
        rasters"""
    raise NotImplementedError


def get_bandnames(rasters):
    """
    Get the filenames and add a number for the bands
    Args:
        rasters: The list of input rasters

    Returns:
            List of rasternames and bandsuffixes
    """
    bandnames = []
    for files in rasters:
        rastername = files.split('/')[-1]
        with rasterio.open(files) as src:
            nr_bands = src.count
            bandnames.extend([f'{rastername}_{x}' for x in
                              np.arange(1, nr_bands+1)])
    return bandnames


def gather_samples_for_roi(roi, rasters, warp_dst):
    """Get sample values from rasters warped to the specified warp destination

        Args:
            roi (fiona shape): region of interest
            rasters (list): list of raster file names
            warp_dst (dict): warp specification

        Returns:
            np.array [nsamples, nfeatures] of samples from within rois.

    """

    try:
        roi_values = []
        for files in rasters:
            with rasterio.open(files) as src:
                with WarpedVRT(
                    src,
                    resampling=Resampling.nearest,
                    width=warp_dst["width"],
                    height=warp_dst["height"],
                    transform=warp_dst["transform"],
                    crs=warp_dst["crs"]
                ) as vrt:
                    selection, _, window = mask.raster_geometry_mask(
                        vrt,
                        [roi["geometry"]],
                        crop=True)
                    raster_subset = vrt.read(window=window)

            # rasterio convention: outside shape=True, inside=False. We invert.
            samples_per_raster = raster_subset[:, ~selection]
            roi_values.append(samples_per_raster)
        roi_values = np.concatenate(roi_values, axis=0).T
    except ValueError:
        DATAPREP_LOGGER.debug("ROI %s OUT OF BOUNDS.. Continuing without "
                              "it", roi['geometry']['coordinates'][0][0])
        roi_values = None
    return roi_values


def gather_samples(rasters, rois, bandnames, out_dir, config_dict):
    """Takes a raster reference from the first raster, loops over all rois and
       returns two lists of equal lenght for samples and labels.s

        Args:
            rasters (list): list of raster file names
            rois (str): path to fiona supported regions of interest file
            bandnames (list): list of names of the input rasters and bands
            out_dir (str): path to the output directory
            config_dict (dict): The cli arguments
        Returns:
            Dataset (DataFrame): Gathered Samples

    """
    sample_labels = []
    all_samples = []
    roi_dict = {}
    if config_dict['su_boxplots']:
        #Boxplot Directory
        roi_bp_dir = out_dir + '/roi_boxplots/'
        os.mkdir(roi_bp_dir)

    # reference everything to first raster
    warp_dst = raster_warp_dst(rasters[0])

    with fiona.open(rois, "r") as shapefile:
        nr_regions = len(shapefile)
        counter = 0
        DATAPREP_LOGGER.debug("%s contains %i regions", rois, nr_regions)
        sample_count = 0
        DATAPREP_LOGGER.info("--Getting Samples--")
        for roi in shapefile:
            progress(counter, nr_regions)
            roi_samples = gather_samples_for_roi(
                roi,
                rasters,
                warp_dst
            )
            if roi_samples is not None:
                sample_count += roi_samples.shape[0]

                all_samples.append(roi_samples.astype(np.float))
                roi_id = int(roi['properties']['id'])
                sample_labels += len(roi_samples) * [roi_id]
                if config_dict['su_boxplots']:
                    polygon_boxplot(roi_samples,
                                    bandnames,
                                    roi_id,
                                    roi_bp_dir,
                                    counter)
                roi_dict[counter] = shape(roi['geometry'])
            counter += 1
        all_samples = np.concatenate(all_samples, axis=0)
        sample_labels = np.asarray(sample_labels)

    all_samples = check_training_data_nodata(all_samples, config_dict)

    dataset = pd.DataFrame(all_samples, columns=bandnames).join(
        pd.DataFrame({'class': sample_labels}))
    DATAPREP_LOGGER.debug("Got: %i samples in total.", len(dataset))

    if config_dict['su_remove_outliers']:
        dataset = outlier_removal(dataset)
    if config_dict['su_boxplots']:
        boxplot(dataset, out_dir)
        folium_map_boxplots(roi_dict, roi_bp_dir)

    return dataset


def createdataset(rasters, rois, workspace, config_dict):
    """
    Create the entire dataset based on rasters and rois
    Args:
        rasters (list): the input rasters
        rois (str): The ogr vector file containing polygons
        workspace (str): Path to workspace
        config_dict (dict): The cli arguments
    Returns:
        dataset
    """
    # Gather the samples and return a full df of all samples
    bandnames = get_bandnames(rasters)
    return gather_samples(rasters,
                          rois,
                          bandnames,
                          workspace,
                          config_dict)

def polygon_boxplot(roi_values, bandnames, class_title, roi_bp_dir, roi_number):
    """Make a simple boxplot for a polygon from the training data

        Args:
            roi_values (np array): A flattened array of roi values for all bands
            bandnames (list): The names of the input bands
            class_title (str): The (landuse) class of the roi
            roi_bp_dir (str): Path where the boxplot figure is saved
            roi_number (int/str): The identifier of the roi to use in filename

    """
    #Bunch of settings to make the boxplots look nicer
    plt.rcParams['xtick.labelsize'] = 4
    plt.rcParams['ytick.labelsize'] = 4
    plt.rcParams['axes.labelsize'] = 4
    plt.rcParams['font.size'] = 4
    plt.rcParams['axes.linewidth'] = 0.2
    plt.rcParams['lines.linewidth'] = 0.2
    plt.rcParams['boxplot.boxprops.linewidth'] = 0.2
    plt.rcParams['boxplot.capprops.linewidth'] = 0.2
    plt.rcParams['boxplot.flierprops.linewidth'] = 0.2
    plt.rcParams['boxplot.flierprops.markersize'] = 2
    plt.rcParams['boxplot.meanprops.linewidth'] = 0.2
    plt.rcParams['boxplot.meanprops.markersize'] = 2
    plt.rcParams['boxplot.medianprops.linewidth'] = 0.2
    plt.rcParams['boxplot.whiskerprops.linewidth'] = 0.2

    sample_df = pd.DataFrame(roi_values, columns=bandnames)
    n_plots = len(sample_df.columns)
    fig, axarr = plt.subplots(1, n_plots)
    for i, columns in enumerate(sample_df.columns):
        subset = sample_df[columns]
        subset.plot(kind='box', ax=axarr[i])
    fig.suptitle(class_title)
    fig.set_size_inches(5, 3)
    plt.tight_layout()
    plt.savefig(os.path.join(roi_bp_dir, f"{roi_number}.png"),
                dpi=300)
    plt.close()

def boxplot(samples, output_dir):
    """Boxplot of all samples

        Args:
            samples (Dataframe): All sample dataset
            output_dir (str): The directory where to save the boxplot

    """
    classes = set(samples['class'])
    n_plots = len(classes)
    if n_plots > 1:
        _, axarr = plt.subplots(n_plots)
        for i, class_number in enumerate(classes):
            subset = samples[samples['class'] == class_number]\
                .drop('class', axis=1)
            subset.boxplot(ax=axarr[i])
            axarr[i].set_title(class_number)
    else:
        _, axis = plt.subplots()
        cols = [x for x in samples.columns if not x == 'class']
        samples[cols].boxplot(ax=axis)
    plt.tight_layout()
    plt.savefig(output_dir+'/all_samples_boxplot.png')


def add_polygons_to_map(gdf, folium_map, png_dir):
    """Adds the separate rois to the map and creates popups with boxplots

        Args:
            gdf (GeoDataFrame): Contains all polygons to add
            folium_map (folium map instance): map to where polygons are added
            png_dir (str): Location of the pngs

        Returns:
            folium map with polygons added

    """
    for items in gdf.iterrows():
        geom_geojson = geojson.Feature(geometry=items[1]['geometry'],
                                       properties={}
                                      )
        points = geom_geojson['geometry']['coordinates']
        image_link = f"{png_dir}/{items[1]['id']}.png"
        with base64.b64encode(open(image_link, 'rb').read()) as src:
            encoded = src.decode()
        html = '<img src="data:image/png;base64,{}">'.format
        iframe = IFrame(html(encoded), width=1600, height=900)
        popup = folium.Popup(iframe, max_width=2560)
        switched_points = [[y, x] for x, y in points[0]]
        marker = folium.PolyLine(switched_points,
                                 radius=10,
                                 popup=popup,
                                 color='#3186cc',
                                 fill=True,
                                 fill_color='#3186cc'
                                )
        folium_map.add_child(marker)
    return folium_map


def add_baselayers(folium_map):
    """Adds the baselayers to the folium map

        Args:
            folium_map(Map): Map to add the baselayers

        Returns:
            folium_map(Map): Map with baselayers

    """
    folium.TileLayer(name='Google Satellite',
                     tiles='https://mt1.google.com/vt/lyrs=s&x={x}&y={y}&z={z}',
                     attr='Google Imagery'
                    ).add_to(folium_map)
    folium.TileLayer(name='Esri Satellite',
                     tiles='https://server.arcgisonline.com/ArcGIS/rest'
                           '/services/World_Imagery/MapServer/tile/{z}/{y}/{'
                           'x}',
                     attr='Esri Imagery').add_to(folium_map)
    folium.LayerControl().add_to(folium_map)
    return folium_map

def folium_map_boxplots(roi_dict, out_dir):
    """Main function to make folium map with boxplots

        Args:
            roi_dict (dict): Roi names and geometries
            out_dir (str): Output directory

    """
    indices = list(roi_dict.keys())
    geoms = list(roi_dict.values())
    gdf = gpd.GeoDataFrame({'id': indices, 'geometry':geoms})
    center_point = MultiPoint(gdf.centroid).centroid
    lat, lon = center_point.y, center_point.x
    folium_map = folium.Map(location=[lat, lon],
                            zoom_start=9,
                           )
    folium_map = add_baselayers(folium_map)
    folium_map = add_polygons_to_map(gdf, folium_map, out_dir)
    folium_map.save(out_dir+'/all_rois.html')

def outlier_removal(samples):
    """Outlier removal from samples using Isolation Forest

        Args:
            samples (DataFrame)

        returns:
            samples (DataFrame)
    """
    DATAPREP_LOGGER.info("Now checking for outliers")

    cols = [x for x in samples.columns if x != 'class']
    if_model = IsolationForest(contamination='auto')

    samples['inlier'] = 0
    lulc_classes = samples['class'].unique()

    # Temporarily here because of annoying warnings in scikit-learn module
    # Needs to be removed after updating scikit-learn to 0.22.0

    with warnings.catch_warnings():
        warnings.simplefilter(action='ignore', category=FutureWarning)
        warnings.simplefilter(action='ignore', category=DeprecationWarning)

        for lulc_class in lulc_classes:
            samples.loc[samples['class'] == lulc_class, 'inlier'] = \
                if_model.fit_predict(samples[samples['class'] == lulc_class][
                    cols])

    # get the number of removed pixels and give output
    class_counts(samples)
    cols_to_return = [x for x in samples.columns if not x == 'inlier']

    return samples[samples['inlier'] > 0][cols_to_return]



def class_counts(samples):
    """Prints the number of pixels removed after outlier removal

        Args:
            samples (DataFrame): Gathered samples with 'inlier' column

    """
    classes = samples['class'].unique()
    count_dict = {}
    for lulc_class in classes:
        count_dict[lulc_class] = [
            samples[samples['class'] == lulc_class].count()[0],
            samples[(samples['class'] == lulc_class) &
                    (samples['inlier'] < 1)].count()[0]
        ]
    # print the header of the table
    to_print = (
        f"\t\t{'Class':<4}"
        f"\t{'#Original':<10}"
        f"\t{'Removed':<10}"
        f"{'Remaining fraction':<10}"
    )
    DATAPREP_LOGGER.info("%s", to_print)

    for class_label, pixel_count in list(count_dict.items()):
        original, removed = pixel_count
        # PRint the row of the table
        to_print = (
            f"\t\t{class_label:<4}"
            f"\t{original:<10}"
            f"\t{removed:<10}"
            f"\t{(1 - removed/original):<10.2f}"
        )
        DATAPREP_LOGGER.info("%s", to_print)



"""All timeseries functionality can be found here"""
import datetime
import logging
from multiprocessing import Manager

import fiona
import numpy as np
import pandas as pd

from classifier.dataprep import gather_samples_for_roi
from classifier.utils.general import impute_values, parallel_function
from classifier.utils.raster import get_raster_date, raster_warp_dst, \
    verify_and_count_bands

TIMESERIES_LOGGER = logging.getLogger(__name__)



def gather_samples_ts_per_roi(
        mp_list, roi, ts_dict, config_dict, warp_dst,
        df_index, bands):
    """Gather samples for a timeseries for a roi, for usage in MP function

        Args:
            mp_list (mp.List): list for use in multiprocessing
            roi (shapely.geometry): region of interest
            ts_dict(dict): dictionary of dates and rasters
            config_dict(dict): Configuration Dictionary
            warp_dst (dict): warp parameters for raster
            df_index (list): index to use for creation of samples df
            bands (list): bands to use from the tifs
    """

    roi_samples = gather_samples_for_roi(
        roi,
        list(ts_dict.values()),
        warp_dst
    )

    index = pd.MultiIndex.from_product(
        [df_index, bands],
        names=['Date', 'Band']
    )
    roi_id = int(roi['properties']['id'])
    roi_classes = roi_samples.shape[0] * [roi_id]
    last_df = pd.DataFrame(
        roi_samples.transpose(),
        index=index,
        columns=roi_classes)

    imputed_df = impute_timeseries(last_df, config_dict)
    mp_list.append(imputed_df)


def gather_samples_ts(ts_dict, config_dict, rois_file=None, roi=None):
    """Gathers the pixel values for all the timeseries rasters and
    combines them in a df

    The DataFrame is a multiColumn dataframe where column level 0 is the class

        Args:
            ts_dict (dict): timeseries raster dictionary
            config_dict(dict): Configuration dictionary
            rois_file (str): Path to the ogr rois file when doing sample
                             gathering for training a model
            roi (shapely.geometry: geometry when gathering pixel values when
                                    predicting a single chunk

        returns:
            samples_df (DataFrame): DataFrame with timeseries of samples
    """

    # Get pandas index
    df_index = [datetime.datetime.strptime(x, "%Y-%m-%d") for x in ts_dict]
    # Iterate through the polygons and return
    rasters = list(ts_dict.values())
    warp_dst = raster_warp_dst(rasters[0])
    b_count = verify_and_count_bands(rasters)
    bands = [f'B{str(x).zfill(2)}' for x in np.arange(1, b_count + 1)]
    with Manager() as manager:
        mp_list = manager.list()
        if rois_file is not None:
            with fiona.open(rois_file, "r") as shapefile:
                args = [
                {
                    'mp_list': mp_list,
                    'roi': roi,
                    'ts_dict': ts_dict,
                    'config_dict': config_dict,
                    'warp_dst': warp_dst,
                    'df_index': df_index,
                    'bands': bands
                }
                    for roi in shapefile
                ]
        else:
            args = [
                {
                    'mp_list': mp_list,
                    'roi': roi,
                    'ts_dict': ts_dict,
                    'config_dict': config_dict,
                    'warp_dst': warp_dst,
                    'df_index': df_index,
                    'bands': bands
                }
            ]
        parallel_function(
            gather_samples_ts_per_roi,
            args,
            ncpus=config_dict['app_threads']
        )
        samples_df = pd.concat(mp_list, axis=1)
        samples_df.columns = samples_df.columns.astype(str)
        return samples_df


def impute_timeseries(samples_df, config_dict):
    """Imputation of timeseries

        Args:
            samples_df(pd.DataFrame): Dataframe with timeseries samples
            config_dict(dict): Configuration dicitionary

        Returns:
            data_filled_df(pd.DataFrame): Imputed DataFrame
    """
    if config_dict['app_imputation_strategy'] == 'interpolate':
        initial_columns = samples_df.columns
        samples_df.columns = [str(x) for x in range(len(samples_df.columns))]

        data_filled_df = samples_df.unstack(level=-1).interpolate(
            method='time',
            axis=0
            ).stack(dropna=False)
        data_filled_df.columns = initial_columns

        if data_filled_df.isnull().values.any():
            # If there are still any nodata pixels left...
            TIMESERIES_LOGGER.warning("Found pixels without any data. "
                                      "Setting mean chunk/roi values "
                                      "to the pixels")

            mean = data_filled_df.stack().mean()
            data_filled_df.fillna(mean, inplace=True)

    else:
        data_array = samples_df.values
        data_filled = [impute_values(data_array, config_dict)]
        data_filled_df = pd.DataFrame(
            np.concatenate(data_filled),
            index=samples_df.index,
            columns=samples_df.columns)
    return data_filled_df


def get_timeseries_samples(rasters, rois, out_dir, config_dict):
    """Samples for timeeseries from a list of rasters and a polygon file

        Args:
            rasters(list): list of raster files
            rois(str): Path to OGR polygon file
            out_dir (str): Path to save the samples to
            config_dict(dict): Configuration dictionary

        Returns:
            samples_df (pd.DataFrame): samples and classes
    """
    raster_dates = [get_raster_date(raster) for raster in rasters]
    raster_dict = dict(zip(raster_dates, rasters))
    samples_df = gather_samples_ts(raster_dict,config_dict, rois)
    if out_dir is not None:
        samples_df.to_csv(f'{out_dir}/samples_ts.csv')
    return samples_df



"""Vector utilities"""
import logging

import geopandas as gpd
import rasterio
from rasterio.features import shapes
from shapely.geometry import Polygon, MultiPolygon

UTILS_VECTOR_LOGGER = logging.getLogger(__name__)

def vectorize_array(array_to_vectorize, transform=None):
    """Vectorize a tif file and return geodataframe.

    Args
        array_to_vectorize (path):a rray to vectorize_and_classify

    Returns:
        geodataframe with polygons

    """
    polygonized = shapes(array_to_vectorize,
                         connectivity=8,
                         transform=transform)
    shapelist, valuelist = zip(*polygonized)
    shapelist = [Polygon(x['coordinates'][0]) for x in shapelist]
    new_poly_dict = {'geometry': shapelist, 'segment_id': valuelist}

    return gpd.GeoDataFrame.from_dict(new_poly_dict)


def get_subset_of_polygons(window, transform, polygons, col_name='max'):
    """Get a subset of a geodataframe from a rasterio window

    Uses a rasterio window to make an intersection on the whole polygon
    dataframe.

    Args:
        window (rasterio.Window): window of interest
        transform (dict):   transform of the window
        polygons (GeoDataFrame): polygons to classify
        col_name (str): Name of the column to subset

    Returns:
        list of [geometry, value] pairs that can be consumed by GeoPandas


    """
    left, bottom, right, top = rasterio.windows.bounds(window,
                                                       transform)

    window_polygon = Polygon([[left, bottom],
                              [left, top],
                              [right, top],
                              [right, bottom]])

    window_df = gpd.GeoDataFrame({'geometry': [window_polygon],
                                  'name': [1]})
    window_df.crs = {'init': 'epsg:4326'}

    subset_polygons = gpd.overlay(window_df,
                                  polygons,
                                  how='intersection')

    return list(zip(list(subset_polygons['geometry'].values),
                    list(subset_polygons[col_name].values)))


def gdf_polygon_to_multipolygon(gdf):
    """Converts all geometries to multipolyon, so there are no issues when
    saving these to a file and adds a crs (epsg:4326) to the gdf

        Args:
            gdf (GeoDataFrame): Containing polygons

        Returns:
            gdf (GeoDataFrame): GDF with all polygons converted to multipolygons

    """
    geometry = [MultiPolygon([feature]) if isinstance(feature, Polygon)
                else feature for feature in gdf["geometry"]]
    return geometry



"The command line interface and entry point for classifier"
import logging
import os

import click

from classifier.accuracy import assess_accuracy
from classifier.classify import classify_, classify_timeseries
from classifier.utils.general import cli_init, save_dict_as_json
from classifier.settings import WORKSPACE

CLI_LOGGER = logging.getLogger(__name__)


@click.group()
def cli():
    """---Classifier---
     Classification of rasters using ground truth"""


@cli.command()
@click.argument(
    'rasters',
    nargs=-1,
    type=click.Path(),
)

@click.option(
    '--rois',
    default=None,
    type=click.Path(),
    help="""OGR supported polygon file"""
)
@click.option(
    '--name',
    type=click.STRING,
    help="""Name of the output folder"""
)

@click.option(
    '--overwrite',
    default=False,
    type=click.BOOL,
    help="""Overwrite Existing output folder"""
)

@click.option(
    '--config_location',
    default=os.path.join(WORKSPACE, 'config.json'),
    type=click.Path(),
    help="""Use a custom location for configuration file"""
)
def classification(name, rasters, rois, overwrite, config_location):
    """Do a traditional (un)supervised classification

    Raster input paths can be added like arguments without a prefix.
    Multiple raster paths can be added

    For additional options, create a config file in your workspace. Also see
    https://satelligence.gitlab.io/classifier/usage/configfile.html.

    """
    out_dir, rasters, rois, config_dict = \
        cli_init(
            name,
            rasters,
            overwrite,
            config_location,
            rois
        )
    config_dict['name'] = name
    if config_dict['app_rasters_are_timeseries']:
        classify_timeseries(rasters, config_dict, out_dir, rois)
    else:
        classify_(rasters, config_dict, out_dir, rois)



@cli.command()
@click.option(
    '--name',
    type=click.STRING,
    help="""Name of the output folder"""
)
@click.option(
    '--raster',
    default=None,
    type=click.Path(),
    help="""Raster file with classification result"""
)
@click.option(
    '--rois',
    default=None,
    type=click.Path(),
    help="""OGR supported polygon file"""
)
@click.option(
    '--subset',
    default=None,
    type=click.Path(),
    help="""OGR supported polygon file"""
)
@click.option(
    '--overwrite',
    default=False,
    type=click.BOOL,
    help="""Overwrite Existing output folder"""
)
@click.option(
    '--config_location',
    default=os.path.join(WORKSPACE, 'config.json'),
    type=click.Path(),
    help="""Use a custom location for configuration file"""
)
def accuracy_assessment(name, raster, rois, subset, overwrite, config_location):
    """Accuracy assessment.
    """
    out_dir, rasters, _, _= cli_init(
        name,
        [raster],
        overwrite,
        config_location)

    cm_fn = os.path.join(out_dir, 'confusion_matrix_total')
    assess_accuracy(rasters[0], rois, cm_fn, subset)


@cli.command()
def make_config():
    """Make a config file with default values in your workspace directory
    and exit"""
    save_dict_as_json()


if __name__ == '__main__':
    cli()



"Testing module of classifier"
import json
import os

from click.testing import CliRunner
import numpy as np
import pytest
import rasterio

from classifier.cli import classification
from classifier.settings import WORKSPACE

@pytest.fixture(scope="module")
def runner():
    """Click Cli Runner for tests"""
    return CliRunner()

def make_config(**kwargs):
    """Make a config file for a specific test
    Args:
        The parameters to change as named variables

    """
    with open(
        os.path.join(WORKSPACE, 'config.json'),
        'w',
        encoding='UTF-8'
    ) as dst:
        dst.write(json.dumps(kwargs))

def setup():
    """
    Setup for the tests
    Returns:

    """

    nowhere = 0.0
    pixel_size = 2.0

    transform = rasterio.Affine(
        pixel_size,
        nowhere,
        nowhere,
        nowhere,
        -pixel_size,
        nowhere
    )

    raster = os.path.join(WORKSPACE, 'test_input.tif')
    data = np.array([
        [
            [1, 200],
            [200, 1]
        ],
        [
            [4, 110],
            [100, 3]
        ]
    ], dtype=np.uint8)

    with rasterio.open(
        raster,
        'w',
        driver='GTiff',
        height=data.shape[1],
        width=data.shape[2],
        count=data.shape[0],
        dtype='uint8',
        crs='+proj=latlong',
        transform=transform
    ) as dst:
        dst.write(data)

    return raster


# pylint:disable=redefined-outer-name
def test_unsupervised(runner):
    """
    Test for running a simple unsupervised classification
    Returns:

    """
    raster = setup()
    make_config(app_threads=1)
    runner.invoke(classification, ["--name", 'test', raster])
    directory = os.path.join(WORKSPACE, 'test')
    with rasterio.open(
        os.path.join(directory, 'classification.tif'),
        'r',
    ) as dst:
        result = dst.read(1)
        assert result[0][0] == 0
        assert result[0][1] == 1
